/***********************************************************************
 *                                                                     *
 *                          Position Editing                           *
 *                                                                     *
 ***********************************************************************/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "gfx.h"
#include "ptrbuf.h"
#include "strbuf.h"
#include "expr.h"
#include "widget.h"
#include "edit_pos.h"

static void eError(char *msg)
{
  gfx_set_text(WwpeMESSAGE,msg);
}

static int getevent(gfx_event_t *event)
{
  while(1) if (gfx_event(NULL,event)) return event->type;
}

static void free_null_set(void *a, void *V, void (*f)(void*))
{
  void **P;
  P= a;
  if (*P) (*f)(*P);
  *P= V;
}

static void free_expr(void *a)
{
  xpp_free_expr(a);
}

static int check_three(int *lab, int p1,int p2,int p3, char *msg)
{
  char *s1,*s2,*s3;
  int E;
  s1= gfx_get_text(lab[p1]);
  s2= gfx_get_text(lab[p2]);
  s3= gfx_get_text(lab[p3]);
  if (s1&&s2&&s3&&s1[0]&&s2[0]&&s3[0]) E= 1; else E= 0;
  free_null_set(&s1, NULL, free);
  free_null_set(&s2, NULL, free);
  free_null_set(&s3, NULL, free);
  if (E) { eError(msg); gfx_focus(lab[p1]); }
  return E;
}

static int check_expr(int ent,char **Rstr, expr_t **Rexpr)
{
  char *S,*M;
  expr_t* E;
  S= gfx_get_text(ent);
  if (!S || !S[0]) { *Rstr= NULL; *Rexpr= NULL; return 0; } // it's cool 
  E= xpp_run(xpp,S);
  if (!E) { M= "Invalid expr"; goto fail; }
  if (E->op==XPP_ERROR) { M= E->data; goto fail; } 

  *Rstr= S;
  *Rexpr= E;
  return 0;

fail:
  eError(M); 
  free(S); 
  if (E) xpp_free_expr(E); 
  gfx_focus(ent); 
  return 1;
}




static int check_uniq(Uscreen_t *screen,Uwidget_t *W,char *str)
{
  Uwidget_t *K;
  for(K=screen->widgets;K;K=K->next)
   if (K!=W && !strcmp(K->ident, str)) return 1;
  return 0;
}

static void choose_var(Uscreen_t *S, int wid, int kind)
{
#if 0
  int n,R;
  Uvar_t *V;
  n= 0;
  for(V=S->vars;V;V=V->next) if (V->vartype==kind) n++;

  if (!n) return ;
  TUPiis_t *ent;
  ent= malloc(sizeof(*ent)*n);
  n= 0;
  for(V=S->vars;V;V=V->next) 
  {
    if (V->vartype==kind)
    { ent[n].iv1= n; ent[n].iv2= 0; ent[n].sv= V->ident; n++; }
  }
  R= gfx_fwmenu(1,n,ent);
  if (R>=0) 
    gfx_text_entry_set(GFX_UPDATE, wid, ent[R].sv);
  free(ent);
#endif
}


static int edit_pos
    (Uscreen_t *screen,char** Rident,Upos_t *opos,
     int inifoc,Uwidget_t* refw,char *title)
{
  char *ident;
  char *pos_str[6];
  expr_t *pos_expr[6];
  int i;
  static int entries[6]=
  {
    [wTOP] WwpeTOP, [wBOT] WwpeBOTTOM, [wLEF] WwpeLEFT,
    [wRIG] WwpeRIGHT, [wWID] WwpeWIDTH, [wHEI] WwpeHEIGHT
  };
  gfx_event_t event;

  gfx_SCRwidpos();

  gfx_set_text(WwpeTITLE, title);
  gfx_set_text(WwpeIDENT, *Rident);

  for(i=0;i<6;i++) gfx_set_text(entries[i], opos->S[i]);

  if (inifoc) gfx_focus(WwpeTOP);
         else gfx_focus(WwpeIDENT);

again:

  while(1)
   switch(getevent(&event))
   {
   case GFXEV_QUIT:
   case WwpeCANCEL:   goto done;
   case WwpeOK:       goto apply;
   case WwpeTOPvb:    choose_var(screen, WwpeTOP,    VAR_Y); break;
   case WwpeBOTTOMvb: choose_var(screen, WwpeBOTTOM, VAR_Y); break;
   case WwpeHEIGHTvb: choose_var(screen, WwpeHEIGHT, VAR_H); break;
   case WwpeLEFTvb:   choose_var(screen, WwpeLEFT,   VAR_X); break;
   case WwpeRIGHTvb:  choose_var(screen, WwpeRIGHT,  VAR_X); break;
   case WwpeWIDTHvb:  choose_var(screen, WwpeWIDTH,  VAR_W); break;
   }
  
done:
   gfx_leave_screen();
   return 0;

apply:
   ident= gfx_get_text(WwpeIDENT);
   if (!ident || !ident[0])
   { 
      eError("Invalid identifier"); 
      gfx_focus(WwpeIDENT);
      goto again;
   }
   if (refw && check_uniq(screen,refw,ident))
   {
     eError("A widget with the same name already exists.");
     gfx_focus(WwpeIDENT); 
     goto again;
   }

   if (check_three(entries, wTOP, wBOT, wHEI, 
                 "Choose two of three: Top, Bottom, Height"))
      goto again;
   
   if (check_three(entries, wLEF, wRIG, wWID,
                  "Choose two of three: Left, Right, Width"))
      goto again;
   
   for(i=0;i<6;i++)
      if (check_expr(entries[i], &pos_str[i], &pos_expr[i])) goto again;

   free_null_set( Rident, ident, free);
   for(i=0;i<6;i++)
   {
      free_null_set(&opos->S[i], pos_str[i], free);
      free_null_set(&opos->E[i], pos_expr[i], free_expr);
   }
   gfx_leave_screen();
   return 0;
}


int edit_widget_position(Uscreen_t *screen,Uwidget_t *W,int inifoc)
{
  return edit_pos(screen, &W->ident, &W->pos,inifoc, W,"Edit Widget Position");
}

int edit_screen_position(Uscreen_t *screen, int inifoc)
{
  return edit_pos(screen, &screen->ident, &screen->pos, inifoc, NULL, 
     "Edit Screen Coordinates");
}

