/***********************************************************************
 *                                                                     *
 *                            LAYOUT EDITOR                            *
 *                                                                     *
 ***********************************************************************
 This is the main file for the layout editor. 
 ***********************************************************************/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include "gfx.h"
#include "ptrbuf.h"
#include "expr.h"
#include "widget.h"
#include "save.h"
#include "edit_pos.h"
#include "show_vars.h"
#include "show_resources.h"
#include "eval.h"
#include "gen.h"
#include "path.h"


#define DISPW 1000
#define DISPH 800


xpp_t *xpp;

resource_t *resources;

static char *cfile,*hfile;
static char *inputf;
static int load_ok;
static int canvas_w, canvas_h;
static int init_canvas_w=800, init_canvas_h=700;
static int canvas_click_mode;
static char *file_type_scr[]= { "Layout Files (*.scr)", NULL };
static char *file_type_c[]= { "C Source Files (*.c)", NULL };
static char *file_type_h[]= { "C Header Files (*.h)", NULL };

static void exchint(int *A,int *B)
{
  int T;
  T= *A;
  *A= *B;
  *B= T;
}

static int get_event(gfx_event_t *event,int *q)
{
  if (*q) return 0;
  while(!gfx_event(NULL,event)) ;
  return 1;
}

static void set_click_mode(int v, char *msg)
{
  canvas_click_mode= v;
  gfx_set_text(Wsta, msg);
}

static void init_expression_parser()
{
  xpp= xpp_new();

  xpp_binary_operator(xpp, "+", EXP_ADD, 50, 'L');
  xpp_binary_operator(xpp, "-", EXP_SUB, 50, 'L');
  xpp_binary_operator(xpp, "*", EXP_MUL, 60, 'L');
  xpp_binary_operator(xpp, "/", EXP_DIV, 60, 'L');
  xpp_binary_operator(xpp, "%", EXP_MOD, 60, 'L');
  xpp_prefix_operator(xpp, "-", EXP_NEG);
  xpp_prefix_operator(xpp, "+", EXP_PLUS);
  xpp_parenthesis(xpp, "(", ")", EXP_PAREN_R);
  xpp_binary_operator(xpp, ".", EXP_DOT, 200, 'L');
}

static int bool_arg(int *argc,char **argv,char *flag, int *Rv)
{
  int i;
  if (*argc<2 || strcmp(argv[1],flag)) return 0;
  for(i=1;i<*argc-1;i++)
    argv[i]= argv[i+1];
  *Rv= 1;
  (*argc)--;
  return 1;
}

static int str_arg(int *argc,char **argv,char *flag, char **Rv)
{
  int i;
  if (*argc<3 || strcmp(argv[1],flag)) return 0;
  *Rv= argv[2];
  for(i=1;i<*argc-2;i++)
    argv[i]= argv[i+2];
  (*argc)-= 2;
  return 1;
}

static unsigned char *rotate_text(unsigned char* d, int w, int h)
{
  unsigned char *c;
  int x,y;
  c= malloc(w*h);
  for(y=0;y<h;y++)
   for(x=0;x<w;x++)
   {
     c[h*x+(h-1-y)]= d[w*y+x];
   }
  free(d);
  return c;
}

static void draw_widget(Uwidget_t *W,int fill)
{
  int x,y,w,h;
  unsigned char *td;
  int tw, th;
  int sx, sy, dx, dy, cw, ch;
  
  x= W->pos.Dx; y= W->pos.Dy; w= W->pos.Dw; h= W->pos.Dh;

  if (fill)
    gfx_canvas_fill_rectangle(Wcan, 0xFFffFF, x,y,w,h);
  gfx_canvas_draw_rectangle(Wcan, 0x000000, x,y,w,h);

  gfx_text_data(1, W->ident, &tw, &th, NULL, &td);

  if (tw>w && h>w)
  {
    td= rotate_text(td,tw,th);
    exchint(&tw,&th);
  }

  if (tw>w) { sx= (tw-w)/2; cw= w;  dx= 0; }
       else { sx= 0;        cw= tw; dx= (w-tw)/2; }
  if (th>h) { sy= (th-h)/2; ch= h;  dy= 0; }
       else { sy= 0;        ch= th; dy= (h-th)/2; }

  gfx_canvas_put_graymap(Wcan, 0x000000, x+dx, y+dy, cw, ch,
                         td, tw, th, sx, sy);
  free(td);
}

static void draw_everything()
{
  Uwidget_t *W;
  gfx_canvas_fill_rectangle(Wcan,0xFFffFF,0,0,canvas_w,canvas_h);
  for(W=screen->widgets;W;W=W->next) 
    draw_widget(W,0); 
} 

static int set_canvas_size(int w,int h)
{
  if (w<=0) w= canvas_w;
  if (h<=0) h= canvas_h;
  if (w==canvas_w && h==canvas_h) return 0;
  canvas_w= w;
  canvas_h= h;
  gfx_canvas_setsize(Wcan,canvas_w,canvas_h);
  return 1;
}

static void handle_canvas_release(gfx_event_t *event)
{
  Uwidget_t *w,**p;
  int x1,y1,x2,y2;
  Uklass_t *k;

  switch(canvas_click_mode)
  {
  case 0: return ;
  case WwidPOS:
     if (!(w= find_widget_bydpos(event->x,event->y))) return ;
     if (edit_widget_position(screen,w,1)) 
       draw_widget(w,1); 
     break;
  case WwidDEL:
     if (!(w= find_widget_bydpos(event->x,event->y))) return ;
     for(p=&screen->widgets;*p && *p!=w;p=&((*p)->next))  ;
     if (!*p) return ;
     *p= w->next; 
     free_widget(w);
     draw_everything();
     break;
  case WwidEDI:
     if (!(w= find_widget_bydpos(event->x,event->y))) return ;
     edit_attribs(w);
     break;
  case WwidNEW:
     gfx_canvas_recsel_begin(Wcan, event); 
     set_click_mode(-canvas_click_mode, 
           "Now click to create the other corner.");
     return ;
  case -WwidNEW:
     gfx_canvas_recsel_end(Wcan, &x1,&y1,&x2, &y2);
     if (x1>x2) exchint(&x1,&x2);
     if (y1>y2) exchint(&y1,&y2);
     k= choose_klass();
     if (!k) break;
     w= make_widget(k,x1,y1,x2,y2);
     edit_widget_position(screen,w,0);
     //edit_attribs(w);  FIXME
     draw_widget(w,0); 
     break;
  }
  set_click_mode(0, "");
}

static void wait_screen(void (*f)())
{
  int q;
  gfx_event_t event;
  (*f)();
  q= 0;
  while(get_event(&event,&q))
    if (event.type==GFXEV_QUIT) break;
  gfx_leave_screen();
}

static void show_eval_results()
{
  char *R;
  size_t nunk;
  unknown_t **unk,*U;
  int i,Z;
  gfx_event_t event;
  int quit= 0;

  R= eval_widget_layout(screen, &nunk, &unk);
  if (R) free(R);

  gfx_SCReval();

  if (nunk)
  {
    gfx_insert(WevLIST,-1,"Following could not be computed:",(void*) -1, NULL);
    for(i=0;i<nunk;i++)
    {
      char buf[200];
      snprintf(buf, sizeof(buf), "  %s = %s", unk[i]->target, unk[i]->str);
      buf[sizeof(buf)-1]= 0;
      gfx_insert(WevLIST,-1, strdup(buf), (void*) (size_t) i, NULL);
    }
  }
  else
  {
    gfx_insert(WevLIST,-1, "Success", (void*) -1, NULL);
  }

  while(get_event(&event,&quit))
   switch(event.type)
   {
   case WevLIST: 
     Z= (int) (size_t) event.usr; 
     if (Z<0 || Z>=nunk || !unk[Z]) continue;
     gfx_remove(WevLIST, event.key, NULL);
     U= unk[Z];
     unk[Z]= NULL;
           // if (U->source=='V') { show_vars(screen, U->ptr); }
     if (U->source=='W') 
        edit_widget_position(screen, U->ptr, U->subsource);
     free(U);
     break;
   case WevOK:
   case GFXEV_QUIT:
     quit= 1;
     break;
   }

  for(i=0;i<nunk;i++) if (unk[i]) free(unk[i]); // FIXME: free the fields also
  if (unk) free(unk);
  gfx_leave_screen();
}

static void select_screen()
{
  int N,i;
  int quit;
  Uscreen_t *selected;
  int selected_index;
  gfx_event_t event;
  Uscreen_t *S,**SS,*p,**pp;
 
  SS= get_all_screens(&N);
  selected= NULL;

  gfx_SCRselscr();

  for(i=0;i<N;i++)
    gfx_insert(WsselLIST, -1, SS[i]->ident, SS[i], NULL);
  free(SS);

  quit= 0;
  while(get_event(&event,&quit))
   switch(event.type)
   {
   case GFXEV_QUIT:
   case WsselCANCEL: 
     selected= NULL;
     quit= 1; 
     break;
   case WsselUSE:
     if (selected) quit= 1;
     break;
   case WsselLIST:
     selected= event.usr;
     selected_index= event.key;
     break;
   case WsselDEL:
               // can't remove the last screen.
     if (!selected || N==1) break;
               // can only remove top level screens. subscreens have to
               // be removed from panel attributes
     for(pp=&screens;*pp!=selected && *pp!=NULL;pp= &((*pp)->next)) ;
     if (*pp==NULL) break;
     *pp= selected->next;
     if (selected==screen) screen= NULL;
     gfx_remove(WsselLIST, selected_index, NULL);
     destroy_screen(selected);
     selected= NULL;
     selected_index= 0;
     N--;
     break;
   case WsselNEW:
     S= create_screen();
     for(p=screens;p->next;p=p->next) ;
     p->next= S;
     gfx_insert(WsselLIST, -1, S->ident, S, NULL);
     edit_screen_position(S, 0);
     gfx_set_selected(WsselLIST, N);
     selected_index= N;
     selected= S;
     N++;
     break;
   }

  gfx_leave_screen();

  if (selected && screen!=selected)
  {
    screen= selected;
    draw_everything();
  } 
  else if (!screen)
  {
    screen= screens;
    draw_everything();
  }
}

static char *size_put(int widget, int value)
{
  char buf[40];
  sprintf(buf,"%d",value);
  gfx_set_text(widget,buf);
}

static int get_size(int widget,int defval)
{
  char *text;
  int value;
  text= gfx_get_text(widget);
  if (!text || !text[0]) return defval;
  value= atoi(text);
  free(text);
  if (value<=0) return defval;
  return value;
}

static char *abs_or_relainp(char *val)
{
  if (!val) return NULL;
  if (val[0]=='/') return val;
  if (inputf) 
      return path_absolute(path_child(path_dir(inputf), val));
  return path_absolute(val);
}

static char *outfile_set(char *base,char *def)
{
  if (!base) base= def;
  return abs_or_relainp(base);
}

static char *outfile_get(int widget,char *def)
{
  char *text;
  text= gfx_get_text(widget);
  if (!text || !text[0]) text= def;
  if (!text || !text[0]) return text;
  return abs_or_relainp(text);
}

static int settings_screen(int *Rcw,int *Rch)
{
  int q;
  gfx_event_t event;
  char *fn;

  gfx_SCRset();
  gfx_set_text(WsetSRC, outfile_set(cfile,"gfx.c"));
  gfx_set_text(WsetHDR, outfile_set(hfile,"gfx.h"));
  size_put(WsetWIDTH,canvas_w);
  size_put(WsetHEIGHT,canvas_h);
  q= 0; 
  while(get_event(&event,&q))
   switch(event.type)
   {
   case WsetSRCbut: 
     fn= gfx_select_file(1, gfx_get_text(WsetSRC), file_type_c);
     if (fn) gfx_set_text(WsetSRC, fn);
     break;
   case WsetHDRbut:
     fn= gfx_select_file(1, gfx_get_text(WsetHDR), file_type_h);
     if (fn) gfx_set_text(WsetHDR, fn);
     break;
   case WsetOK:
     cfile= outfile_get(WsetSRC,cfile);
     hfile= outfile_get(WsetHDR,hfile);
     *Rcw= get_size(WsetWIDTH, canvas_w);
     *Rch= get_size(WsetHEIGHT, canvas_h);
     q= 1;
   case GFXEV_QUIT:
   case WsetCANCEL: q++; break;
   }
  gfx_leave_screen();
  return q-1;
}

static void apply_settings(prj_settings_t *SE,int setcanvas)
{
  if (setcanvas && (SE->canvas_w || SE->canvas_h))
    set_canvas_size(SE->canvas_w,SE->canvas_h);

  if (SE->hdr_file)
    hfile= SE->hdr_file[0]=='/' ? strdup(SE->hdr_file) :
              path_absolute(path_child(path_dir(inputf),SE->hdr_file));
  else
    hfile= NULL;

  if (SE->src_file)
    cfile= SE->src_file[0]=='/' ? strdup(SE->src_file) :
              path_absolute(path_child(path_dir(inputf),SE->src_file));
  else
    cfile= NULL;

  printf("hdr: %s\n", SE->hdr_file);
  printf("src: %s\n", SE->src_file);
}

static void eset(char *msg)
{
  static int empty= 1;
  if (!msg[0] && empty) return ;
  empty= (msg[0]==0);
  gfx_set_text(Wsta,msg);
}

static void main_event_loop()
{
  gfx_event_t event;
  Uscreen_t *S;
  resource_t *R;
  char *fn;
  int quit= 0;
  prj_settings_t settings;
  int cw,ch;

  set_click_mode(0,"");
 
  while(get_event(&event,&quit))
  {
    if (event.type>GFXEV_LAST) eset("");
    switch(event.type)
    {
    case GFXEV_QUIT: return ;
    case WwidPOS: 
      set_click_mode(event.type,"Click a widget to modify its position."); 
      break;
    case WwidDEL:
      set_click_mode(event.type,"Click a widget to delete it.");
      break;
    case WwidEDI:
      set_click_mode(event.type,"Click a widget to edit it.");
      break;
    case WwidNEW:
      set_click_mode(event.type,"Click to create the first corner.");
      break;
    case WscrEDI:
      edit_screen_position(screen, 1);
      break;
    case GFXEV(Wcan,BUTTON_RELEASE):
      if (event.button!=1) break;
      handle_canvas_release(&event);
      break;
    case WfilSAV:
      memset(&settings,0,sizeof(settings));
      settings.canvas_w= canvas_w;
      settings.canvas_h= canvas_h;
      if (!inputf || !load_ok)
      {
        inputf= gfx_select_file(1, "", file_type_scr);
        if (!inputf) break;
      }
      if (hfile) settings.hdr_file= path_relative(hfile,inputf);
      if (cfile) settings.src_file= path_relative(cfile,inputf);
      save_file(inputf, screens, resources,&settings);
      load_ok= 1;
      break;
    case WfilLOD:
      fn= gfx_select_file(0, inputf ? inputf : "", file_type_scr);
      if (!fn || !load_file(fn,&S,&R,&settings)) break; 
        //
        // FIXME: free the current set 
        //
      resources= R;
      screens= screen= S;
      inputf= strdup(fn);
      load_ok= 1;
      apply_settings(&settings,1);
      draw_everything(); 
      break; 
    case Wres: show_resources(); break;
    case WlayVAR: show_vars(screen,NULL); break;
    case WlayEVA: show_eval_results(); break;
    case WlayAPP: 
      evaldis_coords(screen,canvas_w,canvas_h);
      draw_everything(); 
      break; 
    case WscrLIS: select_screen(); break;
    case WfilSET:
      if (settings_screen(&cw,&ch))
      {
        if (set_canvas_size(cw,ch))
           draw_everything();
      }
      break;
    case WfilGEN:
      if (!hfile)
      {
        eset("Set header file from settings first.");
        break;
      }
      if (!cfile)
      {
        eset("Set source file from settings first.");
        break;
      }
      generate_code(hfile,cfile);
      eset("Code generated.");
      break;
    }
  }
}

static int load_initial_file(char *prgname,int force)
{
  load_ok= 0;
  prj_settings_t settings;

  if (inputf)
  {
    Uscreen_t *S; resource_t *R;
    if (load_file(inputf, &S,&R,&settings)) 
    {
      resources= R;
      screens= screen= S;
      load_ok= 1;
      apply_settings(&settings,0);
      if (settings.canvas_w>0) init_canvas_w= settings.canvas_w;
      if (settings.canvas_h>0) init_canvas_h= settings.canvas_h;
    }
  }

  if (!load_ok)
  {
    if (force)
    {
      fprintf(stderr,"%s: an input file must be specified\n", prgname);
      return 1;
    }
    screen= screens= create_screen();
  }
  return 0;
}


int main(int argc,char **argv)
{
  int mode;
  init_expression_parser();

  mode= 0;
  inputf= cfile= hfile= NULL;

  while(argc>1)
  {
    if (bool_arg(&argc,argv,"-g",&mode) ||
        str_arg(&argc,argv,"-h", &hfile) ||
        str_arg(&argc,argv,"-c", &cfile))
      continue;
    if (argv[1][0]=='-')
      exit(fprintf(stderr,"usage: %s [-g] [-h header] [-c source] inputfile\n",
                        argv[0]));
    inputf= argv[1];
    argv[1]= argv[0];
    argv++; argc--;
  }

  if (mode)
  {
    if (load_initial_file(argv[0],1))
       return 1;
    if (!cfile) cfile= path_child(path_dir(inputf),"gfx.c");
    if (!hfile) hfile= path_child(path_dir(inputf),"gfx.h");
    generate_code(hfile,cfile);
    return 0;
  }

  if (gfx_init(NULL,DISPW,DISPH,"Layout Editor")) return 1;

  load_initial_file(argv[0],0);

  gfx_SCREEN_0();
  set_canvas_size(init_canvas_w,init_canvas_h);
  draw_everything();
  main_event_loop();
  gfx_leave_screen();
  gfx_close();
  return 0;
}


#if 0


void make_panel_subs(Uwidget_t *W)
{
  int N,i;
  Uscreen_t **SS;
  Uscreen_t *S;
  char buf[200];

  N= atoi(get_field(W,"n_subs","2"));
  if (N<2) N= 2;

  if (N==W->n_subs) return ;
  
  SS= calloc(1,sizeof(*SS)*N);
  if (W->n_subs)
  {
    for(i=0;i<W->n_subs;i++)
      SS[i]= W->subs[i];
    free(W->subs);
  }
  W->n_subs= N;
  W->subs= SS;

  for(i=0;i<W->n_subs;i++)
  {
    S= W->subs[i];
    if (!S)
    {
      S= calloc(1,sizeof(*S));
      sprintf(buf,"%s.%d", W->ident,i);
      S->ident= strdup(buf);
      S->type= 1;
      W->subs[i]= S;
    }
  }
}

#endif
