#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "expr.h"
#include "widget.h"
#include "save.h"

/***********************************************************************
 *                                                                     *
 *                           SAVE/LOAD FILE                            *
 *                                                                     *
 ***********************************************************************
  The save format consists of commands. Strings are encoded in my special
  hex format where digits are from a to p :)
 ***********************************************************************/


#define WORD_STR(XX) #XX
#define WORD_KEY(XX) KEY_##XX
#define KWORD WORD_KEY

#define FOT_WORDS \
   WORD(VAR) \
      WORD(VALUE) WORD(TYPE) \
   WORD(WIDGET) \
      WORD(KLASS) WORD(FNAME) WORD(FVALUE) WORD(SUBS) WORD(ENDSUBS) \
      WORD(p0) WORD(p1) WORD(p2) WORD(p3) WORD(p4) WORD(p5) \
      WORD(DX) WORD(DY) WORD(DW) WORD(DH) \
   WORD(SCREEN) \
      WORD(Sp0) WORD(Sp1) WORD(Sp2) WORD(Sp3) WORD(Sp4) WORD(Sp5) \
   WORD(ENDSCREEN)  \
   WORD(RESOURCE) \
     WORD(RTYPE) WORD(RLENGTH) WORD(RDATA) WORD(ENDRESOURCE) \
   WORD(WORKAREAW) WORD(WORKAREAH) WORD(SOURCEFILE) WORD(HEADERFILE) \
   WORD(END)

enum fot {
#define WORD(X) WORD_KEY(X),
FOT_WORDS
#undef WORD
};

static void fstr(FILE *f, char *str, char *sfx)
{
  unsigned char *U;
  U= (void*) str;
  while(*U) { fprintf(f,"%c%c", (*U>>4)+'A', (*U&15)+'a'); U++; }
  fprintf(f, "%s",sfx);
}

static void safield_s(FILE *f, char *name, char *value)
{
  fprintf(f,"%s ",name);
  fstr(f, value, "\n");
}

static void save_screen(Uscreen_t *S,FILE *f)
{
  Uvar_t *V; Uwidget_t *W;
  int i;

  safield_s(f, "SCREEN", S->ident);
  for(i=0;i<6;i++) if (S->pos.S[i]) 
        { fprintf(f,"Sp%d ",i); fstr(f,S->pos.S[i],"\n"); }

  for(V=S->vars;V;V=V->next)
  {
    safield_s(f, "VAR", V->ident);
    safield_s(f, "VALUE", V->str);
    fprintf(f,"TYPE %d\n",V->vartype);
  }

  for(W=S->widgets;W;W=W->next)
  {
    safield_s(f,"WIDGET", W->ident);
    for(i=0;i<6;i++)
      if (W->pos.S[i]) { fprintf(f,"p%d ",i); fstr(f,W->pos.S[i],"\n"); }
    fprintf(f,"DX %d\nDY %d\nDW %d\nDH %d\n", 
         W->pos.Dx, W->pos.Dy, W->pos.Dw, W->pos.Dh);
    safield_s(f,"KLASS",W->klass->ident);

    field_t *A;
    for(A=W->fields;A;A=A->next)
    {
      safield_s(f,"FNAME", A->name);
      safield_s(f,"FVALUE", A->data);
    }
    if (W->n_subs!=0)
    {
      fprintf(f,"SUBS %d\n",W->n_subs);
      for(i=0;i<W->n_subs;i++)
        save_screen(W->subs[i], f);
      fprintf(f,"ENDSUBS\n");
    }
  }
  fprintf(f,"ENDSCREEN\n");
}

static void save_resource(resource_t *R,FILE *f)
{
  int i,j;

  safield_s(f,"RESOURCE",R->ident);
  fprintf(f,"RTYPE %d\nRLENGTH %d\n",R->type, R->length);
  for(i=0;i<R->length;i+=20)
  {
    fprintf(f,"RDATA ");
    for(j=0;j<20 && i+j<R->length;j++)
       fprintf(f,"%c%c", ((R->data[i+j]>>4)&0xf)+'a', (R->data[i+j]&0xf)+'a');
    fprintf(f,"\n");
  }
  fprintf(f,"ENDRESOURCE\n");
}

void save_file
      (char *fn,Uscreen_t *SCR,resource_t *RES,prj_settings_t *SET)
{
  FILE *f;
  Uscreen_t *S;
  resource_t *R;

  if (!fn) return ;
  f= fopen(fn,"w");
  if (!f) return ;
  for(S=SCR;S;S=S->next) save_screen(S, f);
  for(R=RES;R;R=R->next) save_resource(R,f);
  if (SET->canvas_w) fprintf(f,"WORKAREAW %d\n",SET->canvas_w);
  if (SET->canvas_h) fprintf(f,"WORKAREAH %d\n",SET->canvas_h);
  if (SET->src_file) safield_s(f, "SOURCEFILE", SET->src_file);
  if (SET->hdr_file) safield_s(f, "HEADERFILE", SET->hdr_file);
  fprintf(f,"END\n");
  fclose(f);
}

static int spfx(char *str, char **PF,char **R)
{
  char *K;
  int i;
  int L;
  for(K=str;K[0];K++) if (K[0]==' ') break;
  for(i=0;PF[i];i++)
  {
    L= strlen(PF[i]);
    if (L!=K-str) continue;
    if (!memcmp(str, PF[i],L)) break;
  }
  if (K[0]) K++; 
  *R= K;
  return i;
}



static enum fot command(char *g,char **Ra)
{
  static char *verbs[]= {
#define WORD(X) WORD_STR(X),
    FOT_WORDS
#undef WORD
    NULL};
  return spfx(g,verbs,Ra);
}

static int getcmdline(char *g,int sz,FILE *f)
{
  int i;
  if (!fgets(g,sz,f)) return 0;
  
  g[sz-1]= 0;
  for(i=0;g[i];i++) if (g[i]=='\r' || g[i]=='\n') { g[i]=0; break; }
  return 1;
}



static char *Sarg(char *V)
{
  char *R;
  int L;
  int i;
  L= strlen(V);
  R= malloc(L/2+1);
  for(i=0;i<L/2;i++) R[i]= ((V[2*i]-'A')<<4) | (V[2*i+1]-'a');
  R[i]= 0;
  return R;
}

static int Iarg(char *V)
{
  return atoi(V);
}

static resource_t *load_resource(char *ident,FILE *f)
{
  resource_t *R;
  char      g[2000];
  char      *A;
  int       offset;
  int       i;
  int       L;

  R= calloc(1,sizeof(*R));
  R->ident= ident;

  while(getcmdline(g,sizeof(g),f))
    switch(command(g,&A))
    {
    case KWORD(ENDRESOURCE): return R;
    case KWORD(RTYPE): R->type= Iarg(A); break;
    case KWORD(RLENGTH): R->length= Iarg(A); R->data= malloc(R->length);
                         offset= 0; break;
    case KWORD(RDATA):
       if (!R->data) continue;
       L= strlen(A);
       if (L%2) L--;
       for(i=0;i<L;i+=2)
       {
         if (offset>=R->length) break;
         R->data[offset]= ((A[i]-'a')<<4) | (A[i+1]-'a');
         offset++;
       }
       break;
    default:
       goto done;
    }
done:
  return R;
}

static Uscreen_t *load_screen(char *ident,FILE *f)
{
  Uscreen_t *S, *sub;
  char      g[2000];
  Uvar_t    *V,**Vtgt;
  Uwidget_t *W,**Wtgt;
  field_t   *T,**Ttgt;
  char      *KN;
  char      *A;
  enum fot  m;
  int       count;
  int       i;

  S= calloc(1,sizeof(*S));
  S->ident= ident;

  Vtgt= &S->vars; V= NULL;
  Wtgt= &S->widgets; W= NULL;
  Ttgt= NULL; T= NULL;

restart:

  while(getcmdline(g,sizeof(g),f))
    switch((m= command(g,&A)))
    {
    case KWORD(VAR):
      V= calloc(1,sizeof(*V)); *Vtgt= V; Vtgt= &V->next;
      V->ident= Sarg(A);
      break; 
    case KWORD(VALUE):
      V->str= Sarg(A);
      V->expr= xpp_run(xpp, V->str);
      break; 
    case KWORD(TYPE):
      V->vartype= Iarg(A);
      break;
    case KWORD(WIDGET):
      W= calloc(1,sizeof(*W));
      W->ident= Sarg(A);
      *Wtgt= W;
      Wtgt= &W->next;
      Ttgt= &W->fields;
      break;
    case KWORD(p0): case KWORD(p1): case KWORD(p2):
    case KWORD(p3): case KWORD(p4): case KWORD(p5):
       W->pos.S[m-KWORD(p0)]= Sarg(A);
       W->pos.E[m-KWORD(p0)]= xpp_run(xpp, W->pos.S[m-KWORD(p0)]);
       break;
    case KWORD(DX): W->pos.Dx= Iarg(A); break;
    case KWORD(DY): W->pos.Dy= Iarg(A); break;
    case KWORD(DW): W->pos.Dw= Iarg(A); break;
    case KWORD(DH): W->pos.Dh= Iarg(A); break;
    case KWORD(KLASS):
       KN= Sarg(A);
       W->klass= klasses[0];
       for(i=0;klasses[i];i++)
        if (!strcmp(KN, klasses[i]->ident)) { W->klass= klasses[i]; break; }
       break;
    case KWORD(FNAME):
       T= calloc(1,sizeof(*T));
       T->name= Sarg(A);
       *Ttgt= T;
       Ttgt= &T->next;
       break;
    case KWORD(FVALUE):
       T->data= Sarg(A);
       break;
    case KWORD(Sp0): case KWORD(Sp1): case KWORD(Sp2): 
    case KWORD(Sp3): case KWORD(Sp4): case KWORD(Sp5): 
       S->pos.S[m-KWORD(Sp0)]= Sarg(A);
       S->pos.E[m-KWORD(Sp0)]= xpp_run(xpp, S->pos.S[m-KWORD(Sp0)]);
       break;
    case KWORD(END):
    case KWORD(ENDSCREEN):
    case KWORD(ENDSUBS):
    case KWORD(SCREEN):
    case KWORD(RESOURCE):
    case KWORD(ENDRESOURCE):
    case KWORD(RTYPE):
    case KWORD(RLENGTH):
    case KWORD(RDATA):
       goto done;
    case KWORD(SUBS):
       W->n_subs= Iarg(A);
       W->subs= calloc(W->n_subs,sizeof(W->subs[0]));
       count= 0;
       while(getcmdline(g,sizeof(g),f))
         switch((m=command(g,&A)))
         {
           case KWORD(SCREEN):
             sub= load_screen(Sarg(A),f);
             sub->type= 1;
             if (count<W->n_subs) W->subs[count++]= sub;
             break;
           case KWORD(ENDSUBS):
             goto restart;
           default:
             goto done;
         }
       goto done;
       break;
    }
done:
  return S;
}

int load_file
      (char *fn, Uscreen_t **Rscr, resource_t **Rres,prj_settings_t *Rset)
{
  FILE *f;
  enum fot m;
  char g[2000];
  Uscreen_t *result, **tgt, *S;
  resource_t *res, *R, **Rtgt;
  char *A;

  memset(Rset,0,sizeof(*Rset));

  f= fopen(fn,"r");
  if (!f) return 0;

  tgt= &result;
  result= NULL;

  Rtgt= &res;
  res= NULL;

  while(getcmdline(g,sizeof(g),f))
    switch((m= command(g,&A)))
    {
    case KWORD(SCREEN):
      S= load_screen(Sarg(A),f);
      *tgt= S;
      tgt= &S->next;
      break;
    case KWORD(RESOURCE):
      R= load_resource(Sarg(A),f);
      *Rtgt= R;
      Rtgt= &R->next;
      break;
    case KWORD(WORKAREAW):
      Rset->canvas_w= Iarg(A);
      break;
    case KWORD(WORKAREAH):
      Rset->canvas_h= Iarg(A);
      break;
    case KWORD(SOURCEFILE):
      Rset->src_file= Sarg(A);
      break;
    case KWORD(HEADERFILE):
      Rset->hdr_file= Sarg(A);
      break;
    default:
    case KWORD(END):
      goto done;
    }
done:

  fclose(f);
  *Rscr= result;
  *Rres= res;
  return 1;
}
